import argparse
from functools import partial
from pprint import pformat

from spaghettree import Ok, Result
from spaghettree.adapters.io_wrapper import IOProtocol, IOWrapper
from spaghettree.domain.adj_mat import AdjMat
from spaghettree.domain.entities import EntityCST
from spaghettree.domain.optimisation import (
    get_dwm,
    merge_single_entity_communities_if_no_gain_penalty,
    optimise_communities,
)
from spaghettree.domain.parsing import (
    create_call_tree,
    extract_entities_and_locations,
    filter_non_native_calls,
    pair_exclusive_calls,
)
from spaghettree.domain.processing import (
    add_empty_inits_if_needed,
    convert_to_code_str,
    create_new_filepaths,
    create_new_module_map,
    infer_module_names,
    remap_imports,
    rename_overlapping_mod_names,
)
from spaghettree.domain.visitors import EntityLocation
from spaghettree.logger import logger


def main(src_root: str, *, new_root: str = "", optimise_src_code: bool = False) -> Result:
    io = IOWrapper()
    return run_process(io, src_root, new_root=new_root, optimise_src_code=optimise_src_code)


def run_process(
    io: IOProtocol, src_root: str, *, new_root: str = "", optimise_src_code: bool = False
) -> Result:
    logger.info(f"*** RUNNING `spaghettree` {src_root = } {new_root = } ***")
    src_code = io.read_files(src_root).unwrap()

    ent_and_locs_res = extract_entities_and_locations(src_code)
    entities, location_map = ent_and_locs_res.unwrap()
    entities_res = filter_non_native_calls(entities)
    entities = entities_res.unwrap()
    call_tree = entities_res.and_then(create_call_tree).unwrap()

    if optimise_src_code:
        return optimise_entity_positions(
            io=io,
            entities=entities,
            location_map=location_map,
            call_tree=call_tree,
            src_root=src_root,
            new_root=new_root,
        )

    adj_mat = AdjMat.from_call_tree_no_optimisation(call_tree).unwrap()
    print(  # noqa: T201
        yellow(
            f"Current Directed Weighted Modularity (DWM): {get_dwm(adj_mat.mat, adj_mat.communities): .5f}"
        )
    )

    return Ok(call_tree)


def optimise_entity_positions(  # noqa: PLR0913
    io: IOProtocol,
    entities: dict[str, EntityCST],
    location_map: dict[str, EntityLocation],
    call_tree: dict[str, list[str]],
    src_root: str,
    new_root: str,
) -> Result:
    return (
        AdjMat.from_call_tree(call_tree)
        .and_then(pair_exclusive_calls)
        .and_then(optimise_communities)
        .and_then(merge_single_entity_communities_if_no_gain_penalty)
        .and_then(partial(create_new_module_map, entities=entities))
        .and_then(infer_module_names)
        .and_then(rename_overlapping_mod_names)
        .and_then(remap_imports)
        .and_then(
            partial(
                convert_to_code_str,
                order_map=location_map,
            ),
        )
        .and_then(partial(create_new_filepaths, new_root=new_root or src_root))
        .and_then(add_empty_inits_if_needed)
        .and_then(partial(io.write_files, ruff_root=new_root or src_root))
    )


def yellow(inp_str: str) -> str:
    return f"\033[33m{inp_str}\033[0m"


def cyan(inp_str: str) -> str:
    return f"\033[36m{inp_str}\033[0m"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Process source code from a given root, with optional relocation and optimisation."
    )

    parser.add_argument("src_root", type=str, help="Path to the source root directory.")
    parser.add_argument(
        "--new-root",
        dest="new_root",
        type=str,
        default="",
        help="Optional new root path for output (default: empty, meaning same as src_root if optimisation is enabled).",
    )
    parser.add_argument(
        "--optimise-src-code",
        dest="optimise_src_code",
        action="store_true",
        help="Enable optimisation of the source code.",
    )

    args = parser.parse_args()
    res = main(args.src_root, new_root=args.new_root, optimise_src_code=args.optimise_src_code)
    call_tree = res.unwrap()
    print(cyan(pformat(call_tree)))  # noqa: T201



from __future__ import annotations

from collections.abc import Collection
from enum import Enum, auto
from typing import Protocol, Self, runtime_checkable

import attrs
import libcst as cst
from attrs.validators import instance_of


@runtime_checkable
class EntityCST(Protocol):
    def get_call_tree_entries(self) -> list[str]: ...

    def resolve_calls(self, import_map: dict[str, str], ent_map: dict[str, str]) -> Self: ...

    def filter_native_calls(self, entities: Collection[str]) -> Self: ...

    def resolve_native_imports(self) -> Self: ...

    def add_referenced_imports(self, imports: set[ImportCST]) -> Self: ...


@attrs.define
class ClassCST:
    name: str = attrs.field(validator=[instance_of(str)])
    tree: cst.ClassDef = attrs.field(validator=[instance_of(cst.ClassDef)], repr=False)
    methods: list[FuncCST] = attrs.field(factory=list, validator=[instance_of(list)])
    imports: set[ImportCST] = attrs.field(factory=set)
    bases: list[str] = attrs.field(factory=list)

    def get_call_tree_entries(self) -> list[str]:
        return [call for meth in self.methods for call in meth.calls]

    def resolve_calls(self, import_map: dict[str, str], ent_map: dict[str, str]) -> Self:
        # piggyback here to get the subclass relationships
        self.bases = resolve_calls(self.bases, import_map, ent_map)

        for meth in self.methods:
            meth.resolve_calls(import_map, ent_map)
        return self

    def filter_native_calls(self, entities: Collection[str]) -> Self:
        for meth in self.methods:
            bases_and_calls = self.bases + meth.calls
            meth.calls = [
                call for call in bases_and_calls if call in entities and meth != self.name
            ]
        return self

    def resolve_native_imports(self) -> Self:
        for method in self.methods:
            for call in method.calls:
                call_parts = call.split(".")
                mod_name = ".".join(call_parts[:-1])
                call_name = call_parts[-1]
                self.imports.add(ImportCST(mod_name, ImportType.FROM, call_name, call_name))
        return self

    def add_referenced_imports(self, imports: set[ImportCST]) -> Self:
        for imp in imports:
            for meth in self.methods:
                if imp.as_name in meth.calls or f"{imp.module}.{imp.as_name}" in meth.calls:
                    self.imports.add(imp)


@attrs.define
class FuncCST:
    name: str = attrs.field(validator=[instance_of(str)])
    tree: cst.FunctionDef = attrs.field(validator=[instance_of(cst.FunctionDef)], repr=False)
    calls: list[str] = attrs.field(factory=list, validator=[instance_of(list)])
    imports: set[ImportCST] = attrs.field(factory=set)

    def get_call_tree_entries(self) -> list[str]:
        return self.calls

    def resolve_calls(self, import_map: dict[str, str], ent_map: dict[str, str]) -> Self:
        self.calls = resolve_calls(self.calls, import_map, ent_map)
        return self

    def filter_native_calls(self, entities: Collection[str]) -> Self:
        self.calls = [call for call in self.calls if call in entities and call != self.name]
        return self

    def resolve_native_imports(self) -> Self:
        for call in self.calls:
            call_parts = call.split(".")
            mod_name = ".".join(call_parts[:-1])
            call_name = call_parts[-1]
            self.imports.add(ImportCST(mod_name, ImportType.FROM, call_name, call_name))
        return self

    def add_referenced_imports(self, imports: set[ImportCST]) -> Self:
        for imp in imports:
            if imp.as_name in self.calls or f"{imp.module}.{imp.as_name}" in self.calls:
                self.imports.add(imp)
        return self


@attrs.define(eq=True)
class GlobalCST:
    name: str = attrs.field()
    tree: cst.SimpleStatementLine = attrs.field(repr=False)
    referenced: list[str] = attrs.field(factory=list)
    imports: set[ImportCST] = attrs.field(factory=set)

    def get_call_tree_entries(self) -> list[str]:
        return self.referenced

    def resolve_calls(self, import_map: dict[str, str], ent_map: dict[str, str]) -> Self:
        self.referenced = resolve_calls(self.referenced, import_map, ent_map)
        return self

    def filter_native_calls(self, entities: Collection[str]) -> Self:
        self.referenced = [ref for ref in self.referenced if ref in entities and ref != self.name]
        return self

    def resolve_native_imports(self) -> Self:
        for ref in self.referenced:
            ref_parts = ref.split(".")
            mod_name = ".".join(ref_parts[:-1])
            ref_name = ref_parts[-1]
            self.imports.add(ImportCST(mod_name, ImportType.FROM, ref_name, ref_name))
        return self

    def add_referenced_imports(self, imports: set[ImportCST]) -> Self:
        if self.name.endswith(".__all__"):
            self.imports.update(imports)
            return self

        for imp in imports:
            if imp.as_name in self.referenced:
                self.imports.add(imp)
        return self


class ImportType(Enum):
    FROM = auto()
    IMPORT = auto()


@attrs.define(frozen=True)
class ImportCST:
    module: str = attrs.field(validator=[instance_of(str)], converter=str.lower)
    import_type: ImportType = attrs.field(validator=[instance_of(ImportType)])
    name: str = attrs.field(validator=[instance_of(str)])
    as_name: str = attrs.field(validator=[instance_of(str)])

    def to_str(self) -> str:
        output: list[str] = []
        if self.import_type is ImportType.FROM:
            output.append(f"from {self.module} import {self.name}")
        elif self.import_type is ImportType.IMPORT:
            output.append(f"import {self.module}")

        if self.name != self.as_name:
            output.append(f"as {self.as_name}")
        return " ".join(output) + "\n"


def resolve_calls(
    calls: list[str],
    import_map: dict[str, str],
    ent_map: dict[str, str],
) -> list[str]:
    resolved_calls: list[str] = []
    for call in calls:
        call_parts = call.split(".")

        if resolved_call := import_map.get(call_parts[0]):
            resolved_call_parts = resolved_call.split(".")
            if resolved_call_parts[-1] != call:
                common_removed = ".".join(resolved_call_parts[:-1])
                resolved_calls.append(f"{common_removed}.{call}".strip("."))
            else:
                resolved_calls.append(resolved_call)
        elif resolved_call := ent_map.get(call_parts[0]):
            resolved_calls.append(resolved_call)
        else:
            resolved_calls.append(call)
    return resolved_calls



from __future__ import annotations

import attrs
import libcst as cst

from spaghettree.domain.entities import ClassCST, FuncCST, GlobalCST, ImportCST, ImportType


@attrs.define(frozen=True, eq=True, order=True)
class EntityLocation:
    path: str = attrs.field()
    name: str = attrs.field(eq=False)
    line_no: int = attrs.field()


class MetadataBase(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (cst.metadata.PositionProvider,)


@attrs.define
class OnePassVisitor(MetadataBase):
    module_name: str = attrs.field()
    current_class: str = attrs.field(default="")
    current_func: str = attrs.field(default="")
    current_global: str = attrs.field(default="")
    depth: int = attrs.field(default=0)
    entities: dict = attrs.field(factory=dict)
    locations: dict = attrs.field(factory=dict)
    imports: set[ImportCST] = attrs.field(factory=set)

    def visit_Import(self, node: cst.Import) -> None:  # noqa: N802
        for alias in node.names:
            name = self._resolve_attr(alias.name)
            asname = alias.asname.name.value if alias.asname else name
            self._add_import(name, ImportType.IMPORT, name, asname)

    def visit_ImportFrom(self, node: cst.ImportFrom) -> None:  # noqa: N802
        module = self._resolve_attr(node.module)
        if module is None:
            # skip relative imports
            return

        if isinstance(node.names, cst.ImportStar):
            self._add_import(module, ImportType.FROM, "*", "*")
            return

        aliases = node.names
        if isinstance(aliases, cst.ImportAlias):
            aliases = [aliases]

        for alias in aliases:
            name = self._resolve_attr(alias.name)
            asname = alias.asname.name.value if alias.asname else name
            self._add_import(module, ImportType.FROM, name, asname)

    def visit_IndentedBlock(self, _: cst.IndentedBlock) -> bool | None:  # noqa: N802
        self.depth += 1

    def leave_IndentedBlock(self, _: cst.IndentedBlock) -> bool | None:  # noqa: N802
        self.depth -= 1

    def visit_Assign(self, node: cst.Assign) -> None:  # noqa: N802
        if self.depth != 0:
            return
        for target in node.targets:
            if not isinstance(target.target, cst.Name):
                return
            self.current_global = target.target.value
            scope = self._get_current_scope()
            self.entities[scope] = GlobalCST(scope, node)
            self._record_location(node, self.current_global)

    def leave_Assign(self, _: cst.Assign) -> None:  # noqa: N802
        self.current_global = ""

    def visit_AnnAssign(self, node: cst.AnnAssign) -> None:  # noqa: N802
        if self.depth == 0 and isinstance(node.target, cst.Name):
            self.current_global = node.target.value
            scope = self._get_current_scope()
            self.entities[scope] = GlobalCST(scope, node)
            self._record_location(node, self.current_global)

    def leave_AnnAssign(self, _: cst.AnnAssign) -> None:  # noqa: N802
        self.current_global = ""

    def visit_ClassDef(self, node: cst.ClassDef) -> None:  # noqa: N802
        if self.depth == 0:
            self.current_class = node.name.value
            scope = self._get_current_scope()
            bases = [self._resolve_attr(arg.value) for arg in node.bases if node.bases]
            self.entities[scope] = ClassCST(scope, node, bases=bases)
            self._record_location(node, node.name.value)

    def leave_ClassDef(self, _: cst.ClassDef) -> None:  # noqa: N802
        if self.depth == 0:
            self.current_class = ""

    def visit_FunctionDef(self, node: cst.FunctionDef) -> None:  # noqa: N802
        if self.depth == 0 or self.current_class:
            self.current_func = node.name.value

        scope = self._get_current_scope()
        func_cst = FuncCST(scope, node)

        if self.current_class:
            self.entities[self._get_current_class_scope()].methods.append(func_cst)
        elif self.depth == 0:
            self._record_location(node, node.name.value)
            self.entities[scope] = func_cst

    def leave_FunctionDef(self, _: cst.FunctionDef) -> None:  # noqa: N802
        if self.current_class or self.depth == 0:
            self.current_func = ""

    def visit_Call(self, node: cst.Call) -> None:  # noqa: N802
        scope = self._get_current_scope()

        if self.current_class and self.current_func:
            # add the calls to the last added method
            self.entities[self._get_current_class_scope()].methods[-1].calls.append(
                self._resolve_attr(node.func)
            )
        elif self.current_func:
            self.entities[scope].calls.append(self._resolve_attr(node.func))

    def visit_Name(self, node: cst.Name) -> None:  # noqa: N802
        scope = self._get_current_class_scope() if self.current_class else self._get_current_scope()

        # class attributes/attrs/dataclasses etc before the first method
        if self.current_class or self.current_func:
            for imp in self.imports:
                if imp.as_name == node.value:
                    self.entities[scope].imports.add(imp)

        # methods/funcs exist
        if self.current_func:
            if self.current_class:
                cls_scope = self._get_current_class_scope()

                if node.value not in self.entities[cls_scope].methods[-1].calls:
                    self.entities[cls_scope].methods[-1].calls.append(node.value)

            elif node.value not in self.entities[scope].calls:
                self.entities[scope].calls.append(node.value)

        if self.current_global:
            self.entities[scope].referenced.append(node.value)

    def _get_current_scope(self) -> str:
        ent = ".".join(
            elem for elem in [self.current_class, self.current_func, self.current_global] if elem
        )
        return f"{self.module_name}.{ent}"

    def _get_current_class_scope(self) -> str:
        return f"{self.module_name}.{self.current_class}"

    def _add_import(self, key: str, import_type: ImportType, name: str, as_name: str) -> None:
        self.imports.add(ImportCST(key, import_type, name, as_name))

    def _record_location(self, node: cst.CSTNode, name: str) -> None:
        self.locations[name] = EntityLocation(
            path=self.module_name,
            name=name,
            line_no=self.get_metadata(cst.metadata.PositionProvider, node).start.line,
        )

    def _resolve_attr(self, node: cst.BaseExpression) -> str | None:
        if isinstance(node, cst.Name):
            return node.value
        if isinstance(node, cst.Attribute):
            parent = self._resolve_attr(node.value)
            return f"{parent}.{node.attr.value}" if parent else node.attr.value
        return None



import os
from collections import Counter, defaultdict
from copy import deepcopy

from spaghettree import safe
from spaghettree.domain.adj_mat import AdjMat
from spaghettree.domain.entities import ImportCST
from spaghettree.domain.parsing import EntityCST, cst_to_str
from spaghettree.logger import logger


@safe
def create_new_module_map(
    adj_mat: AdjMat,
    entities: dict[str, EntityCST],
) -> dict[int, list[EntityCST]]:
    new_modules: defaultdict[int, list[EntityCST]] = defaultdict(list)

    for i, mod_name in enumerate(adj_mat.communities):
        ent_name = adj_mat.node_map[i]
        new_modules[mod_name].append(entities[ent_name])

    return dict(new_modules)


@safe
def infer_module_names(
    new_modules: dict[int, list[EntityCST]],
) -> dict[str, list[EntityCST]]:
    logger.debug(f"{new_modules = }")

    renamed_modules: dict[str, list[EntityCST]] = {}

    for contents in new_modules.values():
        if len(contents) > 1:
            names = [".".join(ent.name.split(".")[:-1]) for ent in contents]
            possible_module_names = sorted(
                {(name, names.count(name)) for name in names}, key=lambda x: (-x[1], x[0])
            )
            logger.debug(f"{possible_module_names = }")

            for name, _ in possible_module_names:
                if name not in renamed_modules:
                    mod_name = name
                    break
            else:
                mod_name = f"{possible_module_names[0][0]}.mod_overflow"
        else:
            mod_name = contents[0].name
        renamed_modules[mod_name] = contents

    logger.debug(f"{renamed_modules = }")
    return renamed_modules


@safe
def rename_overlapping_mod_names(
    renamed_modules: dict[str, list[EntityCST]],
) -> dict[str, list[EntityCST]]:
    def rename_mod_name(name: str, renamed_modules: list[str]) -> str:
        name_parts = name.split(".")
        root = name_parts[0]
        dirname = ".".join(name_parts[:-1])

        dirnames = [".".join(m.split(".")[:-1]) for m in renamed_modules]
        dirname_counts = Counter(dirnames)

        if dirname not in renamed_modules and dirname_counts.get(dirname, 0) <= 1:
            name = dirname
        elif dirname in renamed_modules:
            name = ".".join([*name_parts[:-2], "_".join(name_parts[-2:])])

        if "." not in name:
            name = f"{root}.{name}"
        logger.debug(f"{name = }")
        return name.lower()

    mod_names = list(renamed_modules)
    logger.debug(f"{renamed_modules = }")
    logger.debug(f"{mod_names = }")
    return {
        rename_mod_name(name, mod_names): contents for name, contents in renamed_modules.items()
    }


@safe
def remap_imports(
    modules: dict[str, list[EntityCST]],
) -> dict[str, list[EntityCST]]:
    logger.debug(f"{modules = }")

    modules = deepcopy(modules)
    entity_mod_map: dict[str, str] = {
        ent.name: mod_name for mod_name, ents in modules.items() for ent in ents
    }

    for mod_name, ents in modules.items():
        for ent in ents:
            updated_imports: set[ImportCST] = set()

            for imp in ent.imports:
                new_mod = entity_mod_map.get(f"{imp.module}.{imp.name}")
                if new_mod is None:
                    updated_imports.add(imp)
                elif new_mod != mod_name:
                    updated_imports.add(
                        ImportCST(
                            module=new_mod,
                            import_type=imp.import_type,
                            name=imp.name,
                            as_name=imp.as_name,
                        ),
                    )
            ent.imports = updated_imports
            logger.debug(f"{ent = }")
    return modules


@safe
def create_new_filepaths(
    fixed_name_modules: dict[str, list[EntityCST]],
    new_root: str,
) -> dict[str, list[EntityCST]]:
    logger.debug(f"{fixed_name_modules = }")

    def to_filepath(new_root: str, name: str) -> str:
        return os.path.join(os.path.dirname(new_root), name.replace(".", "/") + ".py")

    return {to_filepath(new_root, name): contents for name, contents in fixed_name_modules.items()}


@safe
def convert_to_code_str(
    new_modules: dict[str, list[EntityCST]],
    order_map: dict[str, int],
) -> dict[str, str]:
    logger.debug(f"{new_modules = }")
    logger.debug(f"{order_map = }")

    def get_module_str(mod_contents: list[EntityCST]) -> str:
        imports, code = [], []

        for ent in mod_contents:
            imports.extend([imp.to_str() for imp in ent.imports])
            code.append(cst_to_str(ent.tree))

        return "".join(sorted(set(imports))) + "".join(code)

    return {
        mod_name: get_module_str(sorted(contents, key=lambda x: order_map[x.name.split(".")[-1]]))
        for mod_name, contents in new_modules.items()
    }


@safe
def add_empty_inits_if_needed(modules: dict[str, str]) -> dict[str, str]:
    logger.debug(f"{modules = }")

    modules_with_inits = {}

    for path, contents in modules.items():
        init_path = f"{os.path.dirname(path)}/__init__.py"

        if init_path not in modules:
            modules_with_inits[init_path] = ""

        if not contents.strip() and os.path.basename(path) != "__init__.py":
            logger.debug(f"Skipping {path = } {contents = }")
            # skip empty files
            continue
        modules_with_inits[path] = contents

    return modules_with_inits
